/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.configuration.repositories.util;

import lombok.AllArgsConstructor;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.projects.ProjectsService;
import org.artifactory.projects.ProjectsUtils;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.rest.model.admin.configuration.repository.info.RepositoryInfo;
import org.artifactory.ui.rest.model.admin.configuration.repository.info.RepositoryInfoListFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * @author Aviad Shikloshi
 */
@Slf4j
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
@AllArgsConstructor
@NoArgsConstructor
public class GetRepositoryInfoService implements RestService {

    @Autowired
    private RepositoryService repositoryService;

    @Autowired
    private CentralConfigService configService;

    @Autowired
    private ProjectsService projectsService;

    @Autowired
    private AuthorizationService authorizationService;

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        String repoType = request.getPathParamByKey("repoType");
        String projectName = request.getPathParamByKey("project");
        if (StringUtils.isEmpty(projectName)) {
            projectName = ProjectsService.DEFAULT_PROJECT;
        }
        final String finalProjectName = projectName;

        Function<RepositoryInfo, Boolean> filterProjectAdminRepositories =
                createFilterProjectAdminRepositoriesFunction(authorizationService.isAdmin(), projectName);
        List<RepositoryInfo> repositoryInfoListInfo =
                RepositoryInfoListFactory
                        .createRepositoryInfo(repoType, configService, repositoryService, finalProjectName).stream()
                        .filter(filterProjectAdminRepositories::apply)
                        .collect(Collectors.toList());

        if (ProjectsUtils.isDefaultProject(finalProjectName)) {
            Set<String> repoKeys =
                    repositoryInfoListInfo.stream().map(RepositoryInfo::getRepoKey).collect(Collectors.toSet());
            Map<String, String> repoKeyToProjectMap = projectsService.getRepoKeyToProjectMap(repoKeys, true);
            repositoryInfoListInfo.forEach(repoInfo -> repoInfo.setProject(
                    repoKeyToProjectMap.getOrDefault(repoInfo.getRepoKey(), ProjectsService.DEFAULT_PROJECT)));
        }
        response.iModelList(repositoryInfoListInfo);
    }

    @NonNull
    private Function<RepositoryInfo, Boolean> createFilterProjectAdminRepositoriesFunction(boolean isAdmin,
            String projectName) {
        Function<RepositoryInfo, Boolean> containsFunction;
        if (isAdmin && ProjectsUtils.isDefaultProject(projectName)) {
            containsFunction = (project) -> true;
        } else {
            Set<String> repositories = projectsService.getProjectRepositories(projectName, true);
            containsFunction = repositoryInfo -> repositories.contains(repositoryInfo.getRepoKey());
        }
        return containsFunction;
    }
}
