package org.artifactory.ui.rest.service.builds.buildsinfo.tabs.permissions;

import com.google.common.base.Strings;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.security.BuildAcl;
import org.artifactory.ui.rest.service.common.EffectivePermissionHelper;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.List;

/**
 * Used in Effective Permission tab when the number of permission targets per entity (user/group) is larger than 5.
 * Then we don't show all the permissions, but only when user is requesting it specifically.
 *
 * @author Yuval Reches
 */
@Slf4j
@Component
public class GetBuildEffectivePermissionsByEntityService extends GetBuildEffectivePermissionServiceBase {

    private final EffectivePermissionHelper effectivePermissionHelper;

    @Autowired
    public GetBuildEffectivePermissionsByEntityService(@NonNull AuthorizationService authService,
            @NonNull UserGroupService userGroupService) {
        super(authService, userGroupService);
        this.effectivePermissionHelper = new EffectivePermissionHelper(userGroupService);
    }

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        BuildInfo buildInfo = getAndValidateRequestDataByQueryParams(request, response);
        if (buildInfo == null) {
            return;
        }

        EntityInfo entityInfo = getEntityInfo(request, buildInfo.getName());
        List<String> permissionsByEntity =
                getPermissionsByEntity(entityInfo, buildInfo.getNumber(), buildInfo.getStarted());
        if (permissionsByEntity != null) {
            response.iModel(permissionsByEntity);
        }
    }

    /**
     * Returns all permissions names of a given entity
     */
    private List<String> getPermissionsByEntity(EntityInfo entityInfo, String buildNumber, String buildStarted) {
        List<BuildAcl> buildNameAcls = userGroupService.getBuildPathAcls(entityInfo.buildName, buildNumber,
                buildStarted);
        boolean isGroup = entityInfo.isGroup;
        String entityName = entityInfo.entityName;
        return effectivePermissionHelper.getPermissionByEntity(buildNameAcls, isGroup, entityName);
    }

    private EntityInfo getEntityInfo(ArtifactoryRestRequest request, String buildName) {
        EntityInfo entityInfo = new EntityInfo();
        entityInfo.buildName = buildName;
        entityInfo.entityName = request.getPathParamByKey("username");
        if (Strings.isNullOrEmpty(entityInfo.entityName)) {
            entityInfo.entityName = request.getPathParamByKey("groupname");
            entityInfo.isGroup = true;
        }
        return entityInfo;
    }

    private static class EntityInfo {
        boolean isGroup = false;
        String entityName;
        String buildName;
    }
}
